(in-package #:asdf-status)

(defparameter *keys*
  '(
    (:author      "Author:")
    (:version     "Version:")
    (:description "Description:")
    (:start       "Started at:")
    (:end         "Ended at:")
    (:elapsed-time "Elapsed time (seconds):")
    (:depends-on  "Depends On:"   process-depends-on)
    (:error       "Error:")
    (:latest-file-date "Most recent change:")
    (:latest-file "Most recent file:")
    ))

;;; ---------------------------------------------------------------------------

(defparameter *error-key*
  '(
    ("asdf-403-error" "403 Error" nil)
    ("asdf-404-error" "404 Error" nil)
    ("asdf-timeout-error" "Timeout Error" nil)
    ("asdf-internal-error" "Internal Error" nil )
    ("asdf-general-error" "Other Error" t)
    ("asdf-success" "Success" t)
    ("asdf-untested" "Untested" nil)))

;;; ---------------------------------------------------------------------------

(defparameter *implementation-codes*
  '("A" "B" "C" "D" "E" "F" "G" "H" "I" "J" "K"
    "L" "M" "N" "O" "P" "Q" "R" "S" "T" "U" "V"
    "W" "X" "Y" "Z"))

;;; ---------------------------------------------------------------------------

(defun systems-matrix (systems &key (filter (constantly t)))
  (sort
   (collect-key-value 
    (merge-elements 
     systems
     (lambda (old new)
       (push new old)
       old)
     (lambda (first)
       (list first))
     :filter (lambda (info) 
               (and (funcall filter info)
                    (not (null (getf info :install)))))
     :key (lambda (info) (string-downcase (kl:ensure-string (getf info :install))))
     :test #'equal)
    :transform (lambda (implementation systems)
                 (list implementation 
                       (sort (car systems) #'string-lessp
                             :key (lambda (info) (getf info :implementation))))))
   #'string-lessp
   :key #'car))

;;; ---------------------------------------------------------------------------

(defun all-implementations-for-matrix (matrix)
  ;;?? pricy
  (sort
   (remove-duplicates
    (remove 
     nil
     (flatten
      (collect-elements
       matrix
       :transform (lambda (datum)
                    (collect-elements 
                     (cadr datum)
                     :transform 
                     (lambda (x) (getf x :implementation)))))))
    :test #'string-equal)
   #'string-lessp))

;;; ---------------------------------------------------------------------------

(defun system-summary-counts (systems filter)          
  (sort
   (element-counts
    systems
    :filter filter
    :key (lambda (x) (list (getf x :implementation) 
                           (null (classify-error x))))
    :test #'equal)
   (make-sorter `((caar string-lessp) (cadar string-lessp)))))

;;; ---------------------------------------------------------------------------

(defun generate-asdf-status-website (systems)
  (copy-static-files)
  (generate-status-pages :systems systems)
  (generate-summaries-by-author systems))

;;; ---------------------------------------------------------------------------

(defun generate-status-pages (&key (systems *asdf-systems*)
                                   (filter (constantly t))
                                   (output (path-of-summary-page "index"))
                                   (verbose? t)
                                   (make-backup? t)
                                   (make-detail-pages? t))
  (let* ((matrix (systems-matrix systems :filter filter))
         (all-implementations (all-implementations-for-matrix matrix))
         (summary-counts (system-summary-counts systems filter))
         
         (backup (ensure-unique-filename-with-date output)))
    
    (assert (<= (size all-implementations) (size *implementation-codes*)))
    
    (generate-summary-html 
     output matrix all-implementations summary-counts 
     :links t
     :backup (and make-backup? backup)
     :make-detail-pages? make-detail-pages?)
    (when make-backup?
      (generate-summary-html 
       backup matrix all-implementations summary-counts :links nil))))

;;; ---------------------------------------------------------------------------

(defun all-authors (systems)
  "Returns a list of the authors we can find in systems"
  (collect-elements 
   (make-iterator systems :unique t
                  :transform (lambda (x) (getf x :author))
                  :test #'equal)))

;;; ---------------------------------------------------------------------------

(defun generate-summary-html (output matrix all-implementations summary-counts
                                     &key links backup 
                                     (make-detail-pages? links)
                                     (title "ASDF Systems Summary"))
  (html-page (output)
    (html
     (:HEAD
      (:TITLE (lml-princ title))
      ((:LINK :REL "stylesheet" :TYPE "text/css" :HREF "system-colors.css" :TITLE "combined"))
      ((:LINK :REL "stylesheet" :TYPE "text/css" :HREF "system-shared.css" :TITLE "combined"))
      ((:LINK :REL "stylesheet" :TYPE "text/css" :HREF "system-summary.css" :TITLE "combined"))
      ((:META :HTTP-EQUIV "Content-Type" :CONTENT "text/html; charset=ISO-8859-1")))
     (:BODY
      ((:DIV :CLASS "header")
       ((:SPAN :CLASS "logo") ((:IMG :SRC "lisp-flag-with-hand.gif")))         
       (:H2 (lml-princ title))
       (:H4 "Generated on "
            (lml-princ (metatilities:format-date "%A, %e %B %Y" (get-universal-time))))
       (:H4 ((:A :HREF "index--authors.html" :TITLE "Author index") "Systems by author"))
       
       #+Ignore
       (when backup
         (html
          (:h4 ((:A :HREF (namestring
                           (make-pathname :name (pathname-name backup)
                                          :type (pathname-type backup)))
                    :TITLE "Results from yesteryear")
                "Previous results")))))
      
      (:P "Brought to you by " 
          ((:A :HREF "http://common-lisp.net/project/ait/")
           "ASDF-Install-Tester") ". Remember, when ASDF is happy, you're happy too.")
      
      ((:table :class "table")
       
       (:tr (:td (generate-systems-matrix
                  matrix all-implementations summary-counts 
                  :links links :make-detail-pages? make-detail-pages?))
            (:td (generate-keys all-implementations :links links)))
       
       (:tr ((:td :colspan 2) ((:div :class "big-height")
                               ((:span :class "hide-me") "+"))))
       
       (:tr ((:td :colspan 2) (generate-footer-links))))))))

;;; ---------------------------------------------------------------------------

(defun generate-systems-matrix (matrix all-implementations summary-counts
                                       &key links (make-detail-pages? links))
  (html
   ((:DIV :CLASS "matrix")
    (:table 
     (:tr
      (:td)
      (loop for nil in all-implementations 
            for code in *implementation-codes* do
            (html ((:th :class "implementation-code") (lml-princ code)))))
     
     (loop for (system datum) in matrix
           for system-name = (string-downcase system) do
           (when (unreported-system-p datum)
             (generate-rss system datum))
           (html
            (:tr
             ((:th :class "header-system-name") 
              ((:A :NAME system-name))
              ((:A :HREF (make-cliki-link system)) (lml-princ system-name)))
             (loop for impl in all-implementations do
                   (acond 
                    ((find-if (lambda (x) 
                                (string-equal impl 
                                              (getf x :implementation)))
                              datum)
                     (let ((detail-html (name-of-detail-page system-name impl)))
                       (when make-detail-pages?
                         (generate-system-detail it))
                       (cond ((not (null (classify-error it)))
                              (html ((:td :class (classify-error it))
                                     (if links
                                       (html 
                                        ((:a :href detail-html :title (classify-error it))
                                         ((:div :class "big-height")
                                          ((:span :class "hide-me") "+"))))
                                       (html 
                                        ((:a :title (classify-error it))
                                         ((:div :class "big-height")
                                          ((:span :class "hide-me") "+"))))))))
                             (t
                              (html ((:td :class "asdf-success")
                                     (if links
                                       (html
                                        ((:a :href detail-html :title "Success")
                                         ((:div :class "big-height")
                                          ((:span :class "hide-me") "+"))))
                                       (html 
                                        ((:a :title "Success")
                                         ((:div :class "big-height")
                                          ((:span :class "hide-me") "+")))))))))))
                    (t
                     (html ((:td :class "asdf-untested")
                            ((:div :class "big-height")
                             ((:span :class "hide-me") "+"))))))))))
     
     (:tr ((:td :colspan (1+ (length all-implementations))) 
           ((:div :class "big-height") ((:span :class "hide-me") "+"))))
     
     ;; summarize
     (flet ((sub-total (implementation value)
              (or (second (assoc (list implementation value) summary-counts
                                 :test #'equal))
                  0)))
       (html 
        (:tr
         ((:th :class "header-system-name") "Success")
         (loop for implementation in all-implementations do
               (html
                ((:td :class "status-summary")
                 (lml-princ
                  (format nil "~:D" (sub-total implementation t)))))))
        (:tr
         ((:th :class "header-system-name") "Failure")
         (loop for implementation in all-implementations do
               (html
                ((:td :class "status-summary") 
                 (lml-princ
                  (format nil "~:D" (sub-total implementation nil)))))))
        
        (:tr
         ((:th :class "header-system-name") "Total")
         (loop for implementation in all-implementations do
               (html
                ((:td :class "status-summary") 
                 (lml-princ
                  (format nil "~:D"
                          (+ (sub-total implementation t) 
                             (sub-total implementation nil))))))))))))))

;;; ---------------------------------------------------------------------------

(defun generate-footer-links ()
  (html
   ((:DIV :CLASS "links")
    (:table
     (:tr (:th "Links"))
     (:tr (:td ((:A :href "http://www.cliki.net/") "CLiki")))
     (:tr (:td ((:A :href "http://www.cliki.net/asdf-install") "ASDF-Install")))
     (:tr (:td ((:A :href "http://www.metabang.com/") "metabang.com")))
     (:tr (:td ((:A :href "http://www.somacon.com/p142.php") "CSS Color Chart")))
     ))))

;;; ---------------------------------------------------------------------------

(defun generate-keys (all-implementations &key links simple-only?)
  (html
   ((:DIV :CLASS "key") 
    (:table
     (:tr ((:th :colspan 2 :class "key-title") "Keys"))
     (generate-status-keys links simple-only?)
     (:tr ((:td :colspan 2 :class "big-height") 
           ((:span :class "hide-me") "+"))) 
     (generate-implementation-keys all-implementations)))))

;;; ---------------------------------------------------------------------------

(defun generate-status-keys (links simple-only?)
  (html
   (when links
     (html (:tr ((:th :colspan 2 :class "key-title") 
                 "\(Click matrix for details\)"))))
   
   (:tr ((:th :colspan 2) "Status Codes"))
   (loop for (class description simple?) in *error-key*
         when (or (not simple-only?) simple?) do
         (html 
          (:tr ((:td :class class)
                ((:div :class "big-height")
                 ((:span :class "hide-me") "+")))
               (:td (lml-princ description)))))))

;;; ---------------------------------------------------------------------------

(defun generate-implementation-keys (all-implementations)
  (html
   (:tr ((:th :colspan 2) "Implementations"))
   (loop for implementation in all-implementations 
         for code in *implementation-codes* do
         (html 
          (:tr ((:th :class "implementation-code") (lml-princ code))
               (:td (lml-princ implementation)))))))

;;; ---------------------------------------------------------------------------

(defun generate-system-detail (system-info)
  (let* ((system-name (getf system-info :install))
         (implementation (getf system-info :implementation))
         (output (path-of-detail-page system-name implementation))
         (lisp-output (make-pathname 
                       :name (concatenate 
                              'string (string-downcase system-name) "-"
                              implementation)
                       :type "output"
                       :defaults *data-directory*))
         (*current-system* (getf system-info :install))
         (*current-implementation* implementation))
    (html-page (output)
      (html
       (:HEAD
        (:TITLE (lml-princ system-name) "/" (lml-princ implementation) " [ASDF Details]")
        ((:LINK :REL "stylesheet" :TYPE "text/css" :HREF "system-colors.css" :TITLE "combined"))
        ((:LINK :REL "stylesheet" :TYPE "text/css" :HREF "system-shared.css" :TITLE "combined"))
        ((:META :HTTP-EQUIV "Content-Type" :CONTENT "text/html; charset=ISO-8859-1")))
       (:BODY
        ((:DIV :CLASS "header")
         ((:SPAN :CLASS "logo") ((:IMG :SRC "lisp-flag-with-hand.gif")))
         (:H2 (lml-princ system-name) " on " (lml-princ implementation) " - Details"))
        
        ((:DIV :CLASS "summary")
         (:table 
          (loop for key-info in *keys* do
                (bind:bind (((key name &optional (process-with #'lml2::princ-safe-http))
                                      key-info)
                       (value (getf system-info key)))
                  (when value
                    (html 
                     (:tr
                      (:td (lml-princ name))
                      (:td (funcall process-with value)))))))))
        
        ((:DIV :CLASS "output")
         (:H2 "REPL Output")
         (:pre (cond ((null lisp-output)
                      (html "Unable to find lisp output"))
                     ((probe-file lisp-output)
                      (lml-insert-file lisp-output))
                     (t
                      (html "Unable to find lisp output")))))
        
        ((:DIV :CLASS "footer")
         ((:A :HREF (format nil "~A#~A" (summary-page-name) system-name)) "Back")))))))

;;; ---------------------------------------------------------------------------

(defun process-depends-on (depends-on)
  (let ((first? t))
    (loop for system in depends-on do
          (let ((system-name (typecase system
                               (symbol (symbol-name system))
                               (string system))))
            (html
             (unless first? (html ", "))
             ((:a :href (path-of-detail-page system-name *current-implementation*)) 
              (lml-princ (string-downcase system-name)))))
          
          (setf first? nil))))

;;; ---------------------------------------------------------------------------

(defun make-cliki-link (system)
  (format nil "http://www.cliki.net/~(~A~)" system))

;;; ---------------------------------------------------------------------------

(defun path-of-detail-page (system-name implementation)
  (make-pathname
   :name (ensure-filename-safe-for-os
          (concatenate 'string
                       (string-downcase system-name) "-"
                       implementation))
   :type "html"
   :defaults *output-directory*))

;;; ---------------------------------------------------------------------------

(defun name-of-detail-page (system-name implementation)
  (namestring (pathname-name+type (path-of-detail-page system-name implementation))))

;;; ---------------------------------------------------------------------------

(defun path-of-summary-page (name) 
  (make-pathname :name (ensure-filename-safe-for-os name)
                 :type "html"
                 :defaults *output-directory*))

;;; ---------------------------------------------------------------------------

(defun summary-page-name ()
  "index.html")

;;; ---------------------------------------------------------------------------

(defun error-summary (systems)
  (collect-elements
   systems
   :transform (lambda (info) (summarize-error info))
   :filter (lambda (info) (not (null (summarize-error info))))))
       
;;; ---------------------------------------------------------------------------

(defun summarize-error (system-info)
  (let ((error-string (getf system-info :error))
        (internal-error-string (getf system-info :internal-error))) 
    (cond ((and (null error-string)
                (null internal-error-string))
           (values nil))
          ((and (search "403" error-string :test #'string-equal)
                (search "http" error-string :test #'string-equal))
           "403 Error")
          ((and (search "404" error-string :test #'string-equal)
                (search "http" error-string :test #'string-equal))
           "404 Error")
          ((search "timeout error" error-string :test #'string-equal)
           "asdf-timeout-error")
          ((not (null internal-error-string))
           "asdf-internal-error")
          (t 
           "Other Error"))))

;;; ---------------------------------------------------------------------------

(defun classify-error (system-info)
  (let ((error-string (getf system-info :error))
        (internal-error-string (getf system-info :internal-error)))
    (cond ((and (null error-string)
                (null internal-error-string))
           (values nil))
          ((and (search "403" error-string :test #'string-equal)
                (search "http" error-string :test #'string-equal))
           "asdf-403-error")
          ((and (search "404" error-string :test #'string-equal)
                (search "http" error-string :test #'string-equal))
           "asdf-404-error")
          ((search "timeout error" error-string :test #'string-equal)
           "asdf-timeout-error")
          ((not (null internal-error-string))
           "asdf-internal-error")
          (t 
           "asdf-general-error"))))

(dsc:system-source-directory 'asdf-status)

;;; ---------------------------------------------------------------------------

(defun copy-static-files ()
  (let ((wild-source (make-pathname 
                      :name :wild
                      :type :wild
                      :directory `(,@(pathname-directory (dsc:system-source-directory 'asdf-status)) "static-files")
                      :defaults (dsc:system-source-directory 'asdf-status)))
        (wild-target (make-pathname
                      :name :wild
                      :type :wild
                      :defaults *output-directory*))) 
    (ensure-directories-exist wild-target)
    (iterate-elements
     (directory wild-source)
     (lambda (file)
       (let ((target (translate-pathname file wild-source wild-target)))
         (when target (delete-file target))
         (metatilities:copy-file file target))))))

;;; ---------------------------------------------------------------------------

(defun unreported-system-p (data)
  (find-if (lambda (x) (getf x :unreported?)) data))
   
;;; ---------------------------------------------------------------------------

(defun generate-rss (system data)
  )


