(in-package #:asdf-status)

;;; ---------------------------------------------------------------------------

(defun process-systems-data (&key (verbose? t))
  (let ((current (load-current-data :verbose? verbose?))
        (new (load-new-data :verbose? verbose?)))
    (prog1 
      (save-data (merge-data current new :verbose? verbose?) :verbose? verbose?)
      (save-input :verbose? verbose?))))

;;; ---------------------------------------------------------------------------

(defun load-current-data (&key (root *data-directory*) (verbose? nil))
  (setf *asdf-systems*
        (let ((file (make-pathname :name "systems"
                                   :type "data"
                                   :defaults root)))
          (when verbose?
            (format t "~%Trying to load current data ~A" file))
          (if (probe-file file)
            (with-open-file (in file) (read in nil nil))
            (when verbose?
              (format t "~%  Error: file doesn't exist")))))) 

;;; ---------------------------------------------------------------------------

(defcondition data-error (error)
  ((file nil ir))
  (:report (lambda (c s)
             (format s "Data error in ~A" (file c)))))

;;; ---------------------------------------------------------------------------

(defun load-new-data (&key (root *input-directory*) (verbose? nil))
  (let ((*package* (find-package 'asdf-status)))
    (ensure-directories-exist *data-directory*)
    (ensure-directories-exist *error-directory*)
    (ensure-directories-exist *save-directory*)
    
    ;; load any new stuff
    (remove-if
     #'atom
     (mapcar 
      (lambda (f)
        (when verbose?
          (format t "~%  loading file ~A" f))
        (handler-case
          (let ((it nil))
            (with-open-file (in f)
              (setf it (read in nil nil)))
            (when (data-error-p it)
              ;; moves files to error directory
              (error 'data-error :file f))
            (setf (getf it :file) f)
            (values it))
          
          (error (c) 
                 (format *error-output* "~%~A" c)
                 (move-files-to-directory (make-pathname
                                           :name (pathname-name f)
                                           :type :wild
                                           :defaults f) 
                                          *error-directory*))))
      (directory (make-pathname
                  :name :wild
                  :type "ait"
                  :directory `(,@(pathname-directory root)
                               :wild-inferiors)
                  :defaults root))))))

;;; ---------------------------------------------------------------------------

(defun save-input (&key (root *input-directory*) (verbose? nil))
  (let* ((*package* (find-package 'asdf-status)))
    (ensure-directories-exist *save-directory*)
    
    (mapc 
     (lambda (f)
       (when verbose? 
         (format t "~%Moving data: ~A" f))
       (move-files-to-directory
        (make-pathname
         :name (pathname-name f)
         :type :wild
         :defaults f)
        *save-directory*))
     (directory (make-pathname
                 :name :wild
                 :type "ait"
                 :directory `(,@(pathname-directory root)
                              :wild-inferiors)
                 :defaults root))))
  (values))

;;; ---------------------------------------------------------------------------

(defun merge-data (old-data new-data &key (verbose? nil))
  ;; three cases? never before seen, seen and not interesting, seen and interesting
  ;; here we collect the never before seen ones and also update the seen and interesting ones.
  (let ((new-records 
         (collect-elements 
          new-data
          :filter
          (lambda (datum)
            (let* ((system (getf datum :install))
                   (implementation (getf datum :implementation))
                   (matching (find-if (lambda (other) 
                                        (and (string-equal system (getf other :install))
                                             (string-equal implementation (getf other :implementation))))
                                      old-data)))
              (cond ((not (null matching))
                     (when (interesting-change-p matching datum)
                       ;; replace
                       (when verbose? 
                         (format t "~%Replace: ~A \(~A -> ~A; ~A -> ~A\)"
                                 system (getf matching :start) (getf datum :start)
                                 (classify-error matching) (classify-error datum)))
                       
                       ;; handle properties
                       (doplist (k v matching)
                         (setf (getf matching (form-symbol "OLD-" k)) v
                               (getf matching k) nil))
                       (doplist (k v datum)
                         (setf (getf matching k) v))
                       
                       ;; handle output
                       (copy-file
                        (make-pathname :type "output" :defaults (getf matching :file))
                        (make-pathname 
                         :name (concatenate 'string (pathname-name (getf datum :file)) "-"
                                            (getf datum :implementation))
                         :type "output"
                         :defaults *data-directory*)
                        :if-exists :supersede
                        :if-does-not-exist :ignore))
                     
                     ;; but don't collect
                     (values nil))
                    (t
                     ;; brand new
                     (when verbose? 
                       (format t "~%New    : ~A, ~A" system implementation)
                       
                       ;; handle output
                       (copy-file
                        (make-pathname :type "output" :defaults (getf datum :file))
                        (make-pathname 
                         :name (concatenate 'string (pathname-name (getf datum :file)) "-"
                                            (getf datum :implementation))
                         :type "output"
                         :defaults *data-directory*)
                        :if-exists :supersede
                        :if-does-not-exist :ignore))
                     t)))))))
    (append old-data
            new-records)))
       
;;; ---------------------------------------------------------------------------

(defun interesting-change-p (old-datum new-datum)
  "New datum is interesting if it's classification is different than what we currently have and its date is
later than what we have."
  (and (> (or (parse-time (getf new-datum :start)) 0)
          (or (parse-time (getf old-datum :start)) 0))
       (not (string-equal (classify-error old-datum)
                          (classify-error new-datum)))))
  
;;; ---------------------------------------------------------------------------

(defun save-data (data &key (root *data-directory*) (verbose? nil))
  (let ((file (make-pathname :name "systems"
                             :type "data"
                             :defaults root)))
    (when verbose?
      (format t "~%Saving data to ~A" file))
    (let ((previous-file nil))
      (when (probe-file file)
        (setf previous-file (ensure-unique-filename-with-date file))
        (move-file file previous-file :if-exists :supersede)) 
    
      (with-new-file (out file :print-right-margin 1024)
        (print data out)))))

;;; ---------------------------------------------------------------------------

(defun data-error-p (datum)
  (or (atom datum)
      (not (getf datum :install))
      ;(not (getf datum :implementation))
      (not (getf datum :start))))

;;; ---------------------------------------------------------------------------

;;?? Gary King 2005-12-21: also in regenerate-websites
(defun lml-insert-file (file)
  (if (probe-file file)
      (with-open-file (in file :direction :input)
        (do ((line (read-line in nil 'eof) (read-line in nil 'eof)))
	    ((eq line 'eof))
	  (html (lml2::princ-safe-http line)
                (lml-princ #\newline))))
    (format *trace-output* "Warning: unable to insert LML file ~S" file)))

;;; ---------------------------------------------------------------------------

(defun move-file (from to &key (if-exists :error))
  ;;?? Possibly much more expensive than a file system move...
  (handler-case 
    (progn
      (copy-file from to :if-exists if-exists) 
      (delete-file from))
    (error (c) 
           ;; is this the best way to re-raise?
           (error c))))

;;; ---------------------------------------------------------------------------

(defun move-files-to-directory (pathspec directory)
  (mapc (lambda (file)
          (move-file 
           file 
           (ensure-unique-filename 
            (make-pathname :name (pathname-name file)
                           :type (pathname-type file)
                           :defaults directory))
           :if-exists :supersede))
        (directory pathspec))
  (values))

;;; ---------------------------------------------------------------------------

(defun ensure-unique-filename (pathspec)
  (if (cl-fad:file-exists-p pathspec)
    (loop for count = 0 then (1+ count) 
          for new-name = (make-pathname
                          :name (concatenate 
                                 'string (pathname-name pathspec) "-" (prin1-to-string count))
                          :type (pathname-type pathspec)
                          :defaults pathspec)
          while (cl-fad:file-exists-p new-name)
          finally (return new-name))
    pathspec))

;;; ---------------------------------------------------------------------------

(defun ensure-unique-filename-with-date (pathspec &optional (date (or (file-write-date pathspec)
                                                                      (get-universal-time))))
  (ensure-unique-filename
   (make-pathname
    :name (concatenate 
           'string (pathname-name pathspec) "-" (format-date "%Y-%m-%d-%H-%M" date))
    :type (pathname-type pathspec)
    :defaults pathspec)))

;;; ---------------------------------------------------------------------------

(defun copy-file-to-directory (pathspec directory &key (if-exists :error))
  "Copies the file designated by the non-wild pathname designator pathspec
to the directory designated by the non-wild pathname designator TO.  If
OVERWRITE is true overwrites the file designtated by TO if it exists."
  (let ((destination (make-pathname :name (pathname-name pathspec)
                                    :type (pathname-type pathspec)
                                    :defaults directory)))
    (copy-file pathspec destination :if-exists if-exists)))
